"""Loaders for mason types."""
import enum
import functools
import json
from typing import Any, Dict, Optional

from google.protobuf import json_format
import yaml

from mason import library as _lib
from mason import node
from mason import port
from mason import schema
from mason.proto import config_pb2
from mason.proto import blueprint_pb2
from mason.proto import library_pb2


def _read_content(filename: str) -> str:
    """Reads the content of a file."""
    with open(filename, 'r') as f:
        return f.read()


def _serialize(value: Any) -> Any:
    """Serializes the value for JSON."""
    if isinstance(value, enum.Enum):
        return str(value.value)
    return str(value)


def _create_node(config: blueprint_pb2.Node,
                 parent: node.Node) -> node.Node:
    """Return new node from config."""
    new_node = parent.create(config.type,
                             uid=config.uid,
                             name=config.name,
                             title=config.title)
    for node_config in config.nodes:
        _create_node(node_config, new_node)
    return new_node


def _create_blueprint(config: blueprint_pb2.Blueprint,
                      library: _lib.Library) -> node.Blueprint:
    """Return new blueprint from config."""
    bp_type = library.blueprint_types.get(config.type, node.Blueprint)
    bp = bp_type(name=config.name, library=library)
    for node_config in config.nodes:
        _create_node(node_config, bp)
    for connection in config.connections:
        bp.connect(connection.source, connection.target)
    return bp


def _dump_port_schema(port_schema: port.Port) -> Dict[str, Any]:
    """Dumps port schema to config."""
    if port_schema.default is not None:
        default = json.dumps(port_schema.default, default=_serialize)
    else:
        default = None
    port_config = library_pb2.Port(name=port_schema.name,
                                   type=port_schema.value_type,
                                   direction=port_schema.direction.value,
                                   sequence=port_schema.is_sequence,
                                   map=port_schema.is_map,
                                   choices=port_schema.choices,
                                   default=default)
    return json_format.MessageToDict(port_config)


def _dump_node_schema(node_schema: schema.Schema) -> Dict[str, Any]:
    """Dumps node schema to config."""
    ports = []
    for _, port_schema in sorted(node_schema.ports.items()):
        ports.append(_dump_port_schema(port_schema))
    node_config = library_pb2.Node(name=node_schema.name,
                                   group=node_schema.group,
                                   ports=ports,
                                   signals=list(sorted(node_schema.signals)),
                                   slots=list(sorted(node_schema.slots)))
    return json_format.MessageToDict(node_config)


def _dump_blueprint_schema(bp_schema: schema.Schema) -> Dict[str, Any]:
    """Dumps the blueprint to config."""
    bp_config = library_pb2.Blueprint(group=bp_schema.group,
                                      name=bp_schema.name,
                                      signals=list(sorted(bp_schema.signals)),
                                      slots=list(sorted(bp_schema.slots)))
    return json_format.MessageToDict(bp_config)


def _load_data(filename: str) -> Dict[str, Any]:
    """Loads content from a file."""
    content = _read_content(filename)
    if filename.endswith('.yaml') or filename.endswith('.yml'):
        return yaml.safe_load(content)
    if filename.endswith('json'):
        return json.loads(content)
    raise RuntimeError(f'Unknown file format: {filename}.')


def dump_library(library: Optional[_lib.Library] = None) -> Dict[str, Any]:
    """Dumps library instance to configuration."""
    library = library or _lib.get_default_library()
    config = library_pb2.Library()
    nodes = [_dump_node_schema(node_type.__schema__)
             for _, node_type in sorted(library.node_types.items())]

    blueprints = [_dump_blueprint_schema(bp_type.__schema__)
                  for _, bp_type in sorted(library.blueprint_types.items())]
    config = library_pb2.Library(nodes=nodes, blueprints=blueprints)
    return json_format.MessageToDict(config)


def load_config(filename: str):
    """Loads a configuration file for mason."""
    data = _load_data(filename)
    conf = json_format.ParseDict(data, config_pb2.Config())
    # pylint: disable=no-member
    modules = set(conf.library.modules)
    if conf.library.extends_default:
        modules.update(_lib.DEFAULT_MODULES)
    _lib.DefaultLibrary = functools.partial(_lib.Library,
                                            version=conf.library.version,
                                            modules=list(modules))
    # pylint: enable=no-member


def load_blueprint(
        filename: str,
        library: Optional[_lib.Library] = None) -> node.Blueprint:
    """Loads a blueprint file from disk."""
    data = _load_data(filename)
    bp_config = json_format.ParseDict(data, blueprint_pb2.Blueprint())
    return _create_blueprint(bp_config, library or _lib.get_default_library())



"""Defines logical nodes."""
import enum
from typing import Any

import mason

class Operator(enum.Enum):
    """Defines comparison operation options."""

    Equal = '=='
    NotEqual = '!='
    LessThan = '<'
    LessThanOrEqual = '<='
    GreaterThan = '>'
    GreaterThanOrEqual = '>='


@mason.nodify
def compare(a: Any, b: Any, op: Operator = Operator.Equal) -> bool:
    """Define a node that compares a to b given the operator."""
    if op == Operator.Equal:
        return a == b
    if op == Operator.NotEqual:
        return a != b
    if op == Operator.LessThan:
        return a < b
    if op == Operator.LessThanOrEqual:
        return a <= b
    if op == Operator.GreaterThan:
        return a > b
    if op == Operator.GreaterThanOrEqual:
        return a >= b
    raise RuntimeError(f'Ivalid operation: {op}')

compare.__node__.__schema__.ports['op'].choices = [
    op.value for op in Operator.__members__.values()]


@mason.nodify
def nonzero(a: Any) -> bool:
    """Returns whether or not the item is empty."""
    return bool(a)



"""Expose public mason API."""

from mason import callbacks
from mason import exceptions
from mason import library
from mason import io
from mason import node
from mason import port

try:
    from importlib import metadata  # Python 3.8+
except ImportError:
    import importlib_metadata as metadata  # Python 3.7<

__version__ = metadata.version('mason-framework')

Blueprint = node.Blueprint
Library = library.Library
Node = node.Node
Port = port.Port
PortDirection = port.PortDirection
Signal = callbacks.Signal

dump_library = io.dump_library
get_default_library = library.get_default_library
inport = port.inport
load_blueprint = io.load_blueprint
load_config = io.load_config
nodify = node.nodify
outport = port.outport
slot = callbacks.slot



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: config.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='config.proto',
  package='mason.config',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=b'\n\x0c\x63onfig.proto\x12\x0cmason.config\"D\n\x07Library\x12\x0f\n\x07version\x18\x01 \x01(\x05\x12\x0f\n\x07modules\x18\x02 \x03(\t\x12\x17\n\x0f\x65xtends_default\x18\x03 \x01(\x08\"0\n\x06\x43onfig\x12&\n\x07library\x18\x01 \x01(\x0b\x32\x15.mason.config.Libraryb\x06proto3'
)




_LIBRARY = _descriptor.Descriptor(
  name='Library',
  full_name='mason.config.Library',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='version', full_name='mason.config.Library.version', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='modules', full_name='mason.config.Library.modules', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='extends_default', full_name='mason.config.Library.extends_default', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=30,
  serialized_end=98,
)


_CONFIG = _descriptor.Descriptor(
  name='Config',
  full_name='mason.config.Config',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='library', full_name='mason.config.Config.library', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=100,
  serialized_end=148,
)

_CONFIG.fields_by_name['library'].message_type = _LIBRARY
DESCRIPTOR.message_types_by_name['Library'] = _LIBRARY
DESCRIPTOR.message_types_by_name['Config'] = _CONFIG
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Library = _reflection.GeneratedProtocolMessageType('Library', (_message.Message,), {
  'DESCRIPTOR' : _LIBRARY,
  '__module__' : 'config_pb2'
  # @@protoc_insertion_point(class_scope:mason.config.Library)
  })
_sym_db.RegisterMessage(Library)

Config = _reflection.GeneratedProtocolMessageType('Config', (_message.Message,), {
  'DESCRIPTOR' : _CONFIG,
  '__module__' : 'config_pb2'
  # @@protoc_insertion_point(class_scope:mason.config.Config)
  })
_sym_db.RegisterMessage(Config)


# @@protoc_insertion_point(module_scope)



"""Defines a library to utilize for generating nodes."""
import importlib
import functools
import types
from typing import Any, Dict, Sequence, Union, Type

from mason import node

DEFAULT_MODULES = (
    'mason.nodes.data',
    'mason.nodes.flow',
    'mason.nodes.log',
    'mason.nodes.logic',
    'mason.nodes.math',
)


class Library:
    """Library to handle dynamic loading and creation of nodes."""

    def __init__(self,
                 version: int = 1,
                 modules: Sequence[str] = None):
        self.version = version
        self.blueprint_types: Dict[str, Type[node.Blueprint]] = {}
        self.node_types: Dict[str, Type[node.Node]] = {}
        if modules:
            for module in modules:
                self.load(module)

    def load(self, module: Union[str, types.ModuleType]):
        """Loads a package of nodes into the library."""
        if isinstance(module, str):
            module = importlib.import_module(module)
        self.load_scope(vars(module))

    def load_scope(self, items: Dict[str, Any]):
        """Loads nodes and blueprints from a dictionary."""
        for item in items.values():
            if hasattr(item, '__node__'):
                item = getattr(item, '__node__')

            try:
                is_blueprint = issubclass(item, node.Blueprint)
                is_node = issubclass(item, node.Node)
            except TypeError:
                is_blueprint = False
                is_node = False

            if is_blueprint:
                self.blueprint_types[item.__schema__.uid] = item
            elif is_node and not item.__abstract__:
                self.node_types[item.__schema__.uid] = item

    def register(self, item_type: Type[node.Node]):
        """Register node or blueprint type to this library."""
        if issubclass(item_type, node.Blueprint):
            self.blueprint_types[item_type.__schema__.uid] = item_type
        elif issubclass(item_type, node.Node):
            self.node_types[item_type.__schema__.uid] = item_type
        else:
            raise TypeError(f'Invalid type to register: {item_type}.')


DefaultLibrary = functools.partial(Library, modules=DEFAULT_MODULES)


@functools.lru_cache()
def get_default_library() -> Library:
    """Return the default library."""
    return DefaultLibrary()



"""Defines Node classes."""
import asyncio
import contextlib
import inspect
import uuid
from typing import Any, Callable, Dict, Optional, Set, Sequence, TypeVar, Union
from typing import Generator, Tuple, Type, TYPE_CHECKING

import attr

from mason import callbacks
from mason import exceptions
from mason import port
from mason import schema

if TYPE_CHECKING:
    from mason import library


@attr.s(auto_attribs=True)
class ExecutionContext:
    """Context information when running a flow."""

    args: Dict[str, Any] = None
    state: Dict[str, Any] = None
    results: Dict[str, Any] = None


class NodeMeta(type):
    """Metaclass for nodes to handle port creation and registration."""

    def __new__(mcs, clsname: str, supers: Tuple[Any], attributes: Any):
        """Generates new node type."""
        is_abstract = attributes.setdefault('__abstract__', False)
        if is_abstract:
            return type.__new__(mcs, clsname, supers, attributes)

        model = type.__new__(mcs, clsname, supers, attributes)
        model.__schema__ = schema.generate(model)
        return model


class Node(metaclass=NodeMeta):
    """Base class for defining nodes."""
    __abstract__ = True
    __schema__ = None

    def __init__(self,
                 *,
                 values: Dict[str, Any] = None,
                 library: Optional['library.Library'] = None,
                 name: str = '',
                 nodes: Optional[Sequence['Node']] = None,
                 parent: Optional['Node'] = None,
                 title: str = '',
                 uid: str = ''):
        if type(self).__abstract__:
            raise NotImplementedError(f'{type(self).__name__} is abstract.')

        self.name = name
        self.uid = uid or str(uuid.uuid4())
        self.ports: Dict[str, port.Port] = {}
        self.signals: Dict[str, callbacks.Signal] = {}

        self._library = library
        self._title = title
        self._parent: Optional['Node'] = None
        self._nodes: Set['Node'] = set()
        self._execution_context: Optional[ExecutionContext] = None

        self._init_schema(type(self).__schema__, values or {})

        if parent:
            self.parent = parent

        if nodes:
            for node in nodes:
                node.parent = self

    def __del__(self):
        """Clears the parent / children hierarchy on deletion."""
        self.parent = None

    def __getitem__(self, key: str) -> Union['Node',
                                             port.Port,
                                             callbacks.Signal,
                                             Callable[..., Any]]:
        """Returns an item within the hierarchy."""
        parts = key.split('.')
        curr = self
        for part in parts[:-1]:
            if part == '__self__':
                continue
            curr = curr.find_child(part)
            if not curr:
                raise KeyError(key)
        prop = parts[-1]
        if prop in curr.ports:
            return curr.ports[prop]
        if prop in curr.signals:
            return curr.signals[prop]
        slots = curr.slots
        if prop in slots:
            return slots[prop]
        nodes = curr.nodes
        if prop in nodes:
            return nodes[prop]
        raise KeyError(key)

    def _init_schema(self,
                     my_schema: Optional[schema.Schema],
                     values: Dict[str, Any]):
        """Initializes this node from its schema."""
        if my_schema:
            for port_name, schema_port in my_schema.ports.items():
                try:
                    overrides = {'value': values[port_name]}
                except KeyError:
                    overrides = {}
                self.ports[port_name] = schema_port.evolve(**overrides)

            for signal_name in my_schema.signals:
                self.signals[signal_name] = callbacks.Signal()

    def connect(self, source_path: str, target_path: str):
        """Convenience method to create a connection between the children."""
        source = self[source_path]
        target = self[target_path]
        source.connect(target)

    def create(self, node_type: Union[str, Type['Node']], **props) -> 'Node':
        """Creates a new child node for this node instance."""
        if isinstance(node_type, str):
            node_type = self.library.node_types[node_type]
        if 'name' not in props:
            props['name'] = self.generate_unique_name(node_type)
        return node_type(parent=self, **props)

    def delete(self):
        """Deletes this node from the graph."""
        self.parent = None
        del self

    def disconnect(self, source_path: str, target_path: Optional[str] = None):
        """Disconnects the source from the target, or from all connections."""
        source = self[source_path]
        target = self[target_path] if target_path else None
        source.disconnect(target)

    @contextlib.contextmanager
    def execution_context(self,
                          initial_state: Dict[str, Any],
                          args: Dict[str, Any]) -> ExecutionContext:
        """Creates a new execution context instance for this blueprint."""
        try:
            self._execution_context = ExecutionContext(args=args,
                                                       results={},
                                                       state=initial_state)
            yield self._execution_context
        finally:
            self._execution_context = None

    async def emit(self, signal_name: str, *args: Any) -> None:
        """Helper function to emit a signal by name."""
        await self.signals[signal_name].emit(*args)

    async def gather(self, *names: str) -> Sequence[Any]:
        """Gathers multiple port values."""
        tasks = (self.ports[name].get() for name in names)
        return await asyncio.gather(*tasks)

    def generate_unique_name(self, node_type: Type['Node']) -> str:
        """Returns a unique name for a given base in this container."""
        my_schema = node_type.__schema__
        base = my_schema.name.lower() if my_schema else 'node'
        count = 1
        while True:
            name = f'{base}-{count:02}'
            child = self.find_child(name)
            if not child:
                return name
            count += 1

    async def get(self, name: str) -> Any:
        """Returns the port value for the given name."""
        return await self.ports[name].get()

    def get_context(self) -> Optional[ExecutionContext]:
        """Returns the context for this node."""
        curr = self
        # pylint: disable=protected-access
        while curr:
            if curr._execution_context:
                return curr._execution_context
            curr = curr.parent
        # pylint: enable=protected-access
        return None

    def find_ancestor(self, node_type: Type['Node']) -> Optional['Node']:
        """Returns the ancestor whose type matches the given parameter."""
        curr = self.parent
        while curr and not isinstance(curr, node_type):
            curr = curr.parent
        return curr

    def find_child(
            self,
            name: str,
            recursive: bool = False,
            node_type: Optional[Type['Node']] = None) -> Optional['Node']:
        """Returns a child by its name."""
        for child in self._nodes:
            if ((not node_type or isinstance(child, node_type)) and
                    child.name == name):
                return child
            if not recursive:
                continue
            found = child.find_child(name, recursive=True, node_type=node_type)
            if found:
                return found
        return None

    @property
    def library(self) -> 'library.Library':
        """Traverse up the hierarchy to find the closest library."""
        curr = self
        # pylint: disable=protected-access
        while curr and not curr._library:
            curr = curr.parent
        if not curr:
            from mason import library as _lib  # pylint: disable=import-outside-toplevel
            return _lib.get_default_library()
        # pylint: disable=protected-access
        return curr._library

    @property
    def blueprint(self) -> Optional['Blueprint']:
        """Returns the blueprint node in this hierarchy."""
        return self.find_ancestor(Blueprint)

    @property
    def root_node(self) -> 'Node':
        """Returns the root node in this hierarchy."""
        curr = self
        while curr.parent:
            curr = curr.parent
        return curr

    @property
    def parent(self) -> Optional['Node']:
        """Returns the parent for this node."""
        return self._parent

    @property
    def nodes(self) -> Dict[str, 'Node']:
        """Returns the children for this node."""
        return {n.name: n for n in self._nodes}

    @parent.setter
    def parent(self, parent: Optional['Node']):
        """Sets the parent for this node."""
        if parent == self._parent:
            return
        if self._parent:
            self._parent._nodes.remove(self)  # pylint: disable=protected-access
        self._parent = parent
        if parent:
            parent._nodes.add(self)  # pylint: disable=protected-access

    async def setup(self):
        """Setup this node before execution."""

    @property
    def slots(self) -> Dict[str, Callable[..., Any]]:
        """Returns a mapping of slots for this node."""
        my_schema = type(self).__schema__
        if my_schema:
            return {slot_name: getattr(self, slot_name)
                    for slot_name in my_schema.slots}
        return {}

    @property
    def title(self) -> str:
        """Generates a title for this node."""
        return self._title or (
            self.name
            .title()
            .replace('_', ' ')
            .replace('-', ' ')
        )

    def walk_nodes(self) -> Generator['Node', None, None]:
        """Traverses the hierarchy of this node."""
        for node in self._nodes:
            yield node
            yield from node.walk_nodes()

    async def teardown(self):
        """Teardown this node after execution."""


class Blueprint(Node):
    """Defines a base blueprint node."""

    triggered: callbacks.Signal

    async def __call__(self,
                       initial_state: Dict[str, Any] = None,
                       **args: Any) -> Any:
        """Executes the blueprint."""
        with self.execution_context(initial_state, args) as context:
            for node in self.walk_nodes():
                await node.setup()
            try:
                await self.emit('triggered')
            except exceptions.ReturnException as exc:
                results = exc.value or context.results
            except exceptions.ExitException as exc:
                if exc.code != 0:
                    raise
            else:
                results = context.results
            finally:
                for node in self.walk_nodes():
                    await node.teardown()
            return results if results else None


def nodify(*args,
           name: str = '',
           result_name: str = 'result',
           node_type: Optional[Type[Node]] = None):
    """Decorator for generating a node type from a function."""
    T = TypeVar('T')
    def inner(func: T) -> T:
        sig = inspect.signature(func)
        args = []
        signals = []
        annotations = {}
        defaults = {}
        pass_context_as = ''
        for param_name, param in sig.parameters.items():
            try:
                is_context = issubclass(param.annotation, ExecutionContext)
            except TypeError:
                is_context = False

            if is_context:
                pass_context_as = param_name
            elif param.annotation is callbacks.Signal:
                signals.append(param_name)
                annotations[param_name] = param.annotation
            else:
                args.append(param_name)
                annotations[param_name] = param.annotation
                if param.default is not inspect.Parameter.empty:
                    defaults[param_name] = param.default

        def init(inst: Node, **props):
            Node.__init__(inst, **props)

            try:
                inst.ports[result_name].getter = getattr(inst, func.__name__)
            except KeyError:
                pass

        async def bound_func(inst: Node) -> Any:
            """Method to invoke the function."""
            kwargs = {}
            if pass_context_as:
                kwargs[pass_context_as] = inst.get_context()
            values = await inst.gather(*args)
            kwargs.update(zip(args, values))
            kwargs.update({sig_name: inst.signals[sig_name]
                           for sig_name in signals})
            result = func(**kwargs)
            if inspect.isawaitable(result):
                return await result
            return result

        if hasattr(func, 'is_slot'):
            setattr(bound_func, 'is_slot', func.is_slot)
            setattr(bound_func, 'connection_count', 0)

        if sig.return_annotation is not inspect.Parameter.empty:
            annotations[result_name] = port.Port(
                sig.return_annotation,
                direction=port.PortDirection.Output)

        node_attrs = {
            '__annotations__': annotations,
            '__init__': init,
            func.__name__: bound_func,
            **defaults,
        }

        node_base = (node_type or Node,)
        node_group = inspect.getmodule(func).__name__.split('.')[-1]
        node_clsname = name or func.__name__.title().replace('_', '')
        node_class = type(node_clsname, node_base, node_attrs)
        node_class.__schema__.group = node_group
        func.__node__ = node_class
        return func
    if len(args) == 1:
        return inner(args[0])
    return inner



"""Command line interface for mason."""

import importlib
import os

import click
import mason

_DEFAULT_CLI_PLUGS = {'mason_server.server'}
_MASON_CLI_PLUGS = set(
    filter(bool, os.environ.get('MASON_CLI_PLUGS', '').split(',')))
_MASON_CLI_PLUGS.update(_DEFAULT_CLI_PLUGS)


@click.group()
@click.option('--config', help='Mason config file.', default='')
def cli(config: str = None):
    """Mason command line interface."""
    if config:
        mason.load_config(config)


@cli.command()
def version():
    """Print out the current version."""
    print(mason.__version__)


# Initialize command plugins
for cli_plug in _MASON_CLI_PLUGS:
    try:
        module = importlib.import_module(cli_plug)
    except ImportError:
        if cli_plug not in _DEFAULT_CLI_PLUGS:
            raise



"""Defines data nodes."""
from typing import Any, Optional


import mason


@mason.nodify
def const(default: Any) -> Any:
    """Returns the constant value."""
    return default


class Get(mason.Node):
    """Gets a valu from the execution context state."""

    key: Optional[str] = None
    default: Any = None

    value: mason.outport(Any)

    def __init__(self, **props):
        super().__init__(**props)
        self.ports['value'].getter = self.get_value

    async def get_value(self) -> Any:
        """Returns the value from the context."""
        key, default = await self.gather('key', 'default')
        context = self.get_context()
        if context:
            return context.state.get(key or self.name, default)
        return default


class Set(mason.Node):
    """Sets a value in the execution context state."""

    key: Optional[str] = None
    value: Any = None

    @mason.slot
    async def store(self):
        """Stores the value at the current time to the context state."""
        key, value = await self.gather('key', 'value')
        context = self.get_context()
        if context:
            context.state[key or self.name] = value



"""Defines model schema files."""
import inspect
import typing
from typing import Dict, Set, Type, TYPE_CHECKING

import attr

from mason import callbacks
from mason import port

if TYPE_CHECKING:
    from mason import node


@attr.s(auto_attribs=True, eq=False)
class Schema:
    """Define schema properties."""

    group: str
    name: str
    ports: Dict[str, 'port.Port']
    signals: Set[str]
    slots: Set[str]

    @property
    def uid(self) -> str:
        """Returns the unique id for this schema."""
        return f'{self.group}.{self.name}'



def generate(model: Type['node.Node']) -> Schema:
    """Generates a new Schema instance from the type attributes."""
    group_name = inspect.getmodule(model).__name__.rsplit('.', 1)[-1]
    model_name = model.__name__
    ports = {}
    signals = set()
    slots = set()

    # extract signals and ports from annotations
    for name, annotation in typing.get_type_hints(model).items():
        if name.startswith('_'):
            continue

        if annotation is callbacks.Signal:
            signals.add(name)
        elif isinstance(annotation, port.Port):
            annotation.name = name
            ports[name] = annotation
        else:
            default = getattr(model, name, None)
            ports[name] = port.Port(annotation, name=name, default=default)

    # extract slots from attributes
    for name, prop in inspect.getmembers(model):
        if getattr(prop, 'is_slot', False):
            slots.add(name)

    schema = Schema(group=group_name,
                    name=model_name,
                    ports=ports,
                    signals=signals,
                    slots=slots)
    return schema



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: blueprint.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import any_pb2 as google_dot_protobuf_dot_any__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='blueprint.proto',
  package='mason.blueprint',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=b'\n\x0f\x62lueprint.proto\x12\x0fmason.blueprint\x1a\x19google/protobuf/any.proto\",\n\nConnection\x12\x0e\n\x06source\x18\x01 \x01(\t\x12\x0e\n\x06target\x18\x02 \x01(\t\"9\n\x04Port\x12\x0c\n\x04name\x18\x01 \x01(\t\x12#\n\x05value\x18\x02 \x01(\x0b\x32\x14.google.protobuf.Any\"\x8a\x01\n\x04Node\x12\x0b\n\x03uid\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\r\n\x05title\x18\x03 \x01(\t\x12\x0c\n\x04type\x18\x04 \x01(\t\x12$\n\x05ports\x18\x05 \x03(\x0b\x32\x15.mason.blueprint.Port\x12$\n\x05nodes\x18\x06 \x03(\x0b\x32\x15.mason.blueprint.Node\"\x9b\x01\n\tBlueprint\x12\x0b\n\x03uid\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\r\n\x05title\x18\x03 \x01(\t\x12\x0c\n\x04type\x18\x04 \x01(\t\x12$\n\x05nodes\x18\x05 \x03(\x0b\x32\x15.mason.blueprint.Node\x12\x30\n\x0b\x63onnections\x18\x06 \x03(\x0b\x32\x1b.mason.blueprint.Connectionb\x06proto3'
  ,
  dependencies=[google_dot_protobuf_dot_any__pb2.DESCRIPTOR,])




_CONNECTION = _descriptor.Descriptor(
  name='Connection',
  full_name='mason.blueprint.Connection',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='source', full_name='mason.blueprint.Connection.source', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='target', full_name='mason.blueprint.Connection.target', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=63,
  serialized_end=107,
)


_PORT = _descriptor.Descriptor(
  name='Port',
  full_name='mason.blueprint.Port',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='mason.blueprint.Port.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='mason.blueprint.Port.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=109,
  serialized_end=166,
)


_NODE = _descriptor.Descriptor(
  name='Node',
  full_name='mason.blueprint.Node',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='uid', full_name='mason.blueprint.Node.uid', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='name', full_name='mason.blueprint.Node.name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='title', full_name='mason.blueprint.Node.title', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='type', full_name='mason.blueprint.Node.type', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='ports', full_name='mason.blueprint.Node.ports', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='nodes', full_name='mason.blueprint.Node.nodes', index=5,
      number=6, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=169,
  serialized_end=307,
)


_BLUEPRINT = _descriptor.Descriptor(
  name='Blueprint',
  full_name='mason.blueprint.Blueprint',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='uid', full_name='mason.blueprint.Blueprint.uid', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='name', full_name='mason.blueprint.Blueprint.name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='title', full_name='mason.blueprint.Blueprint.title', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='type', full_name='mason.blueprint.Blueprint.type', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='nodes', full_name='mason.blueprint.Blueprint.nodes', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='connections', full_name='mason.blueprint.Blueprint.connections', index=5,
      number=6, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=310,
  serialized_end=465,
)

_PORT.fields_by_name['value'].message_type = google_dot_protobuf_dot_any__pb2._ANY
_NODE.fields_by_name['ports'].message_type = _PORT
_NODE.fields_by_name['nodes'].message_type = _NODE
_BLUEPRINT.fields_by_name['nodes'].message_type = _NODE
_BLUEPRINT.fields_by_name['connections'].message_type = _CONNECTION
DESCRIPTOR.message_types_by_name['Connection'] = _CONNECTION
DESCRIPTOR.message_types_by_name['Port'] = _PORT
DESCRIPTOR.message_types_by_name['Node'] = _NODE
DESCRIPTOR.message_types_by_name['Blueprint'] = _BLUEPRINT
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Connection = _reflection.GeneratedProtocolMessageType('Connection', (_message.Message,), {
  'DESCRIPTOR' : _CONNECTION,
  '__module__' : 'blueprint_pb2'
  # @@protoc_insertion_point(class_scope:mason.blueprint.Connection)
  })
_sym_db.RegisterMessage(Connection)

Port = _reflection.GeneratedProtocolMessageType('Port', (_message.Message,), {
  'DESCRIPTOR' : _PORT,
  '__module__' : 'blueprint_pb2'
  # @@protoc_insertion_point(class_scope:mason.blueprint.Port)
  })
_sym_db.RegisterMessage(Port)

Node = _reflection.GeneratedProtocolMessageType('Node', (_message.Message,), {
  'DESCRIPTOR' : _NODE,
  '__module__' : 'blueprint_pb2'
  # @@protoc_insertion_point(class_scope:mason.blueprint.Node)
  })
_sym_db.RegisterMessage(Node)

Blueprint = _reflection.GeneratedProtocolMessageType('Blueprint', (_message.Message,), {
  'DESCRIPTOR' : _BLUEPRINT,
  '__module__' : 'blueprint_pb2'
  # @@protoc_insertion_point(class_scope:mason.blueprint.Blueprint)
  })
_sym_db.RegisterMessage(Blueprint)


# @@protoc_insertion_point(module_scope)



"""Defines a basic Port type."""

import enum
import functools
import weakref
from collections import abc
from typing import Any, Awaitable, Callable, Dict, Optional, Set, Sequence
from typing import Union

from mason import exceptions

_CHOICES_TYPE = Optional[Union[Dict[str, Any], Sequence[Any]]]
_GETTER_TYPE = Callable[[], Awaitable[Any]]

class PortDirection(enum.Enum):
    """Defines flow direction types."""

    Input = 'input'
    Output = 'output'


class Port:
    """Defines a data storage class for nodes."""

    def __init__(self,
                 annotation: Any,
                 *,
                 choices: _CHOICES_TYPE = None,
                 default: Any = None,
                 direction: PortDirection = PortDirection.Input,
                 getter: Optional[_GETTER_TYPE] = None,
                 name: str = '',
                 title: str = '',
                 value: Any = None):

        self.annotation = annotation
        self.name = name
        self.default = default
        self.direction = direction
        self.getter = getter

        origin = getattr(annotation, '__origin__', None)

        self._choices = choices
        self.is_map = origin is dict
        try:
            self.is_sequence = issubclass(origin, abc.Sequence)
        except TypeError:
            self.is_sequence = False

        try:
            self.is_enum = issubclass(annotation, enum.Enum)
        except TypeError:
            self.is_enum = False

        self._title = title
        self._connections: Set['Port'] = weakref.WeakSet()

        if isinstance(value, Port):
            self._local_value = default
            self.connect(value)
        else:
            self._local_value = value if value is not None else default

    def __del__(self):
        """Disconnects this port from others on deletion."""
        self.disconnect()

    def can_connect(self, other: 'Port') -> bool:
        """Tests to see whether or not this port can connect to the other."""
        if self.direction == other.direction:
            return False
        if self.direction == PortDirection.Output:
            return True
        return self.is_sequence or self.is_map or not self.is_connected

    @property
    def choices(self) -> _CHOICES_TYPE:
        """Returns an optional sequence of strings for this port."""
        if self._choices:
            return self._choices
        if self.is_enum:
            return dict(self.annotation.__members__)
        return None

    @choices.setter
    def choices(self, value: _CHOICES_TYPE):
        """Sets the choices for this port."""
        self._choices = value

    def connect(self, *ports: 'Port') -> None:
        """Creates a connection between this port and another port."""
        for port in ports:
            if self.can_connect(port) and port.can_connect(self):
                self._connections.add(port)
                port._connections.add(self)  # pylint: disable=protected-access
            else:
                raise exceptions.InvalidConnectionError(self, port)

    def evolve(self, **overrides: Any) -> 'Port':
        """Generate a copy of this port information."""
        props = dict(
            annotation=self.annotation,
            choices=self._choices,
            name=self.name,
            default=self.default,
            direction=self.direction,
            getter=self.getter,
            value=self._local_value,
            title=self._title
        )
        props.update(overrides)
        return Port(**props)

    def disconnect(self, *ports: 'Port') -> None:
        """Disconnects all ports, or the port specified as other."""
        if not ports:
            for port in self._connections:
                port._connections.remove(self)  # pylint: disable=protected-access
            self._connections.clear()
        else:
            for port in ports:
                port._connections.remove(self)  # pylint: disable=protected-access
                self._connections.remove(port)

    async def get(self) -> Any:
        """Returns the current value of this port."""
        is_input = self.direction == PortDirection.Input
        use_connection = is_input and self.is_connected

        if use_connection and self.is_sequence:
            return {await other.get() for other in self._connections}
        if use_connection and self.is_map:
            return {other.name: await other.get()
                    for other in self._connections}
        if use_connection:
            conn = next(iter(self._connections))
            return await conn.get()
        if self.getter:
            return await self.getter()
        return self.local_value

    @property
    def is_connected(self) -> bool:
        """Returns whether or not this port has connections."""
        return len(self._connections) > 0

    @property
    def local_value(self) -> Any:
        """Returns the local value of this port instance."""
        return self._local_value

    @local_value.setter
    def local_value(self, value: Any):
        """Sets the local value for this port instance."""
        if self.is_enum and not isinstance(value, self.annotation):
            if hasattr(self.annotation, value):
                value = getattr(self.annotation, value)
            else:
                value = self.annotation(value)

        self._local_value = value

    @property
    def title(self) -> str:
        """Returns the title of this port."""
        return self._title or self.name.title()

    @property
    def value_type(self) -> Any:
        """Returns the value type of this port."""
        args = getattr(self.annotation, '__args__', None)
        if args:
            vtype = args[-1]
        else:
            vtype = self.annotation
        return getattr(vtype, '__name__', str(vtype))


inport = functools.partial(Port, direction=PortDirection.Input)
outport = functools.partial(Port, direction=PortDirection.Output)



"""Define flow control nodes."""

import asyncio
from typing import Any, Sequence

import mason


@mason.nodify
@mason.slot
def return_(value: Any = None):
    """Raises the value as a return."""
    raise mason.exceptions.ReturnException(value)


@mason.nodify
@mason.slot
def exit_(code: int = 0):
    """Raises the exit exception."""
    raise mason.exceptions.ExitException(code)


class Input(mason.Node):
    """Defines a node to extract an input variable from a flow."""

    key: str
    default: Any
    value: mason.outport(Any)

    def __init__(self, **args):
        super().__init__(**args)

        self.ports['value'].getter = self.get_value

    async def get_value(self) -> Any:
        """Extracts the value for the given output port from the context."""
        key, default = await self.gather('key', 'default')
        if not key:
            key = self.name
        context = self.get_context()
        return context.args.get(key, default) if context else default



class For(mason.Node):
    """Defines a For Loop node."""

    start: int = 0
    stop: int = 10
    interval: int = 1

    index: mason.outport(int)

    index_changed: mason.Signal
    cancelled: mason.Signal
    finished: mason.Signal

    def __init__(self, **props):
        super().__init__(**props)

        self._cancelled = asyncio.Event()

    @mason.slot
    async def run(self):
        """Implement internal run function."""
        self._cancelled.clear()
        start, stop, interval = await self.gather('start', 'stop', 'interval')
        for i in range(start, stop, interval):
            self.ports['index'].local_value = i
            await self.emit('index_changed')
            if self._cancelled.is_set():
                await self.emit('cancelled')
                break
        else:
            await self.emit('finished')

    @mason.slot
    async def cancel(self):
        """Cancels the loop."""
        self._cancelled.set()


class Iterate(mason.Node):
    """Iterate over a set of items."""

    items: mason.inport(Sequence[Any])
    item: mason.outport(Any)

    item_changed: mason.Signal
    cancelled: mason.Signal
    finished: mason.Signal

    def __init__(self, **props):
        super().__init__(**props)
        self._cancelled = asyncio.Event()

    @mason.slot
    async def run(self):
        """Implement internal iteration logic."""
        self._cancelled.clear()
        items = await self.get('items')
        for item in items:
            self.ports['item'].local_value = item
            await self.emit('item_changed')
            if self._cancelled.is_set():
                await self.emit('cancelled')
                break
        else:
            await self.emit('finished')


class Enumerate(mason.Node):
    """Iterate over a set of items."""

    items: Sequence[Any]
    index: mason.outport(int)
    item: mason.outport(Any)

    item_changed: mason.Signal
    cancelled: mason.Signal
    finished: mason.Signal

    def __init__(self, **props):
        super().__init__(**props)
        self._cancelled = asyncio.Event()

    @mason.slot
    async def run(self):
        """Implement internal iteration logic."""
        self._cancelled.clear()
        items = await self.get('items')
        for index, item in enumerate(items):
            self.ports['index'].local_value = index
            self.ports['item'].local_value = item
            await self.emit('item_changed')
            if self._cancelled.is_set():
                await self.emit('cancelled')
                break
        else:
            await self.emit('finished')


class Merge(mason.Node):
    """Wait for all connections before continuing."""

    merged: mason.Signal

    def __init__(self, **props):
        super().__init__(**props)
        self._counter = 0

    async def setup(self):
        """Initialize counter to 0."""
        self._counter = 0

    @mason.slot
    async def continue_(self):
        """Emits finished when all tasks are complete."""
        self._counter += 1
        if self._counter == getattr(self.continue_, 'connection_count', 0):
            await self.emit('merged')
            self._counter = 0


class While(mason.Node):
    """While loop node."""

    condition: Any

    triggered: mason.Signal
    cancelled: mason.Signal
    finished: mason.Signal

    def __init__(self, **props):
        super().__init__(**props)

        self._cancelled = asyncio.Event()

    @mason.slot
    async def run(self):
        """Perform while loop."""
        self._cancelled.clear()
        while await self.get('condition'):
            await self.emit('triggered')
            if self._cancelled.is_set():
                await self.emit('cancelled')
                break
        else:
            await self.emit('finished')

    @mason.slot
    async def cancel(self):
        """Cancels the while loop."""
        self._cancelled.set()


class If(mason.Node):
    """Control logic for if/else checks."""

    condition: Any

    passed: mason.Signal
    failed: mason.Signal

    @mason.slot
    async def check(self):
        """Checks whether or not the condition is true and emits signals."""
        if await self.get('condition'):
            await self.emit('passed')
        else:
            await self.emit('failed')


@mason.nodify
@mason.slot
async def sleep(seconds: float = 1, finished: mason.Signal = None):
    """Control node for sleeping."""
    await asyncio.sleep(seconds)
    if finished:
        await finished.emit()



"""Define a generic callback system."""

import asyncio
import inspect
import types
from typing import Any, Callable, Optional, Set, Union, TypeVar
import weakref

CallbackType = Union[types.MethodType, Callable[..., Any]]
SlotType = TypeVar('SlotType')


class Signal:
    """Signal class type."""

    def __init__(self, *annotations):
        params = []
        for i, annotation in enumerate(annotations):
            params.append(
                inspect.Parameter(
                    f'arg_{i}',
                    inspect.Parameter.POSITIONAL_ONLY,
                    annotation=annotation
                )
            )
        self._annotations = annotations
        self._signature = inspect.Signature(params)
        self._slot_refs = set()

    def _cleanup_dead_refs(self) -> None:
        """Clean out dead references."""
        self._slot_refs = {ref for ref in self._slot_refs if ref()}

    def _get_active_slots(self) -> Set[CallbackType]:
        """Cleans out dead references and returns active callbacks."""
        self._cleanup_dead_refs()
        return {ref() for ref in self._slot_refs}

    def _validate_slot_signature(self, func: CallbackType) -> None:
        """Ensures that the signature of the slot function matches the signal.

        This method will iterate over the annotations for the Signal and compare
        them against the function signature for the callback slot.  If the type
        annotations do not match, a TypeError is raised.

        Args:
            func: Callable object with type annotations.

        Raises:
            TypeError if the signatures are not compatible.
        """
        signature = inspect.signature(func)
        if len(self._annotations) <= len(signature.parameters):
            for i, param in enumerate(signature.parameters.values()):
                actual = param.annotation
                try:
                    expected = self._annotations[i]
                except IndexError:
                    if (param.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD
                            and param.default == inspect.Parameter.empty):
                        break
                else:
                    # Check against the name to support type annotations of
                    # classes that are referenced before being defined.  See
                    # test case for example.
                    name = getattr(actual, '__name__', None)
                    if expected not in (actual, name):
                        break
            else:
                return

        raise TypeError('Failed to create connection.  Expected'
                        f'{self._annotations} but received {signature}.')

    def connect(self, *funcs: CallbackType) -> None:
        """Creates a connection by validating and adding the slot to the ref."""
        for func in funcs:
            self._validate_slot_signature(func)
            if getattr(func, 'is_slot', False) is True:
                func.__dict__.setdefault('connection_count', 0)
                func.__dict__['connection_count'] += 1
            if inspect.ismethod(func):
                ref = weakref.WeakMethod(func)
            else:
                ref = weakref.ref(func)
            self._slot_refs.add(ref)

    def disconnect(self, *funcs: Optional[CallbackType]) -> None:
        """Removes the slot (if provided) or all slot connections."""
        ignore = [None] + list(funcs)
        new_refs = set()
        for ref in self._slot_refs:
            func = ref()
            if func not in ignore:
                if getattr(func, 'is_slot', False) is True:
                    func.__dict__['connection_count'] -= 1
                continue
            new_refs.add(ref)
        self._slot_refs = new_refs

    async def emit(self, *args: Any) -> None:
        """Iterates over each active slot and calls them with the given values.

        Args:
            *args: Variable arguments to send to connected slot functions.

        Raises:
            TypeError if the provided arguments do not match the signature.
        """
        if self._signature.bind(*args):
            tasks = (func(*args) for func in self._get_active_slots())
            await asyncio.gather(*tasks)

    @property
    def is_empty(self) -> bool:
        """Cleans up dead references and returns if any active slots remain."""
        self._cleanup_dead_refs()
        return not self._slot_refs


def slot(func: SlotType) -> SlotType:
    """Decorate the function as a slot."""
    setattr(func, 'is_slot', True)
    setattr(func, 'connection_count', 0)
    return func



"""Define dynamic math functions."""

import math
import functools
import mason

math_node = functools.partial(mason.nodify, result_name='value')

@math_node
def add(a: float, b: float) -> float:  # pylint: disable=C0103
    """Return the result of adding a and b."""
    return a + b

@math_node
def sub(a: float, b: float) -> float:  # pylint: disable=C0103
    """Return the result of subtracting b from a."""
    return a - b

@math_node
def mult(a: float, b: float) -> float:  # pylint: disable=C0103
    """Return the product of multiplication of a and b."""
    return a * b

@math_node
def div(dividend: float, divisor: float) -> float:
    """Return the result of division of the dividend by the divisor."""
    return dividend / divisor

@math_node
def neg(number: float) -> float:
    """Return the negative of number."""
    return number * -1

@math_node
def pow_(base: float, exponent: float = 2) -> float:  # pylint: disable=C0103
    """Return the exponentiation of base by exponent."""
    return base ** exponent

@math_node
def sqrt(number: float) -> float:
    """Return the square root of number."""
    return math.sqrt(number)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mason', 'mason.nodes', 'mason.proto']

package_data = \
{'': ['*']}

install_requires = \
['attrs>=19.3.0,<20.0.0',
 'click>=7.1.2,<8.0.0',
 'importlib-metadata>=1.6.0,<2.0.0',
 'protobuf>=3.11.3,<4.0.0',
 'pyyaml>=5.3.1,<6.0.0']

entry_points = \
{'console_scripts': ['mason = mason.cli:cli']}

setup_kwargs = {
    'name': 'mason-framework',
    'version': '0.0.5',
    'description': 'Node based building blocks.',
    'long_description': None,
    'author': 'Eric Hulser',
    'author_email': 'eric.hulser@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""Define logging nodes."""
import logging
from typing import Any, Optional

import mason


class Print(mason.Node):
    """Prints out a message."""

    message: Optional[Any] = None

    @mason.slot
    async def print_(self):
        """Prints current message."""
        print(await self.get('message'))


class Log(mason.Node):
    """Logs out to the base python logger."""

    name: str = 'root'
    level: str = 'INFO'
    message: Any = None

    @mason.slot
    async def log(self):
        """Logs to the logger."""
        name, level, message = await self.gather('name', 'level', 'message')
        log_level = getattr(logging, level)
        logger = logging.getLogger(name)
        logger.log(log_level, message)



"""Defines exception classes for the mason framework."""
from typing import Any, TYPE_CHECKING

if TYPE_CHECKING:
    from mason import port


class MasonError(Exception):
    """Base error for all Mason exceptions."""


class FlowException(Exception):
    """Base exception to signal the exit of a flow."""


class ExitException(FlowException):
    """Exception to exit a flow with a code."""

    def __init__(self, code: int, message: str = ''):
        self.code = code
        super().__init__(message)


class ReturnException(FlowException):
    """Exception to return a value from the flow."""

    def __init__(self, value: Any):
        self.value = value
        super().__init__()


class InvalidConnectionError(MasonError):
    """Raised when two ports cannot connect to one another."""

    def __init__(self, a: 'port.Port', b: 'port.Port'):
        super().__init__(f'Cannot connect {a.name} to {b.name}.')



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: library.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='library.proto',
  package='mason.library',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=b'\n\rlibrary.proto\x12\rmason.library\"v\n\x04Port\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0c\n\x04type\x18\x02 \x01(\t\x12\x11\n\tdirection\x18\x03 \x01(\t\x12\x10\n\x08sequence\x18\x04 \x01(\x08\x12\x0b\n\x03map\x18\x05 \x01(\x08\x12\x0f\n\x07\x63hoices\x18\x06 \x03(\t\x12\x0f\n\x07\x64\x65\x66\x61ult\x18\x07 \x01(\t\"g\n\x04Node\x12\r\n\x05group\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\"\n\x05ports\x18\x03 \x03(\x0b\x32\x13.mason.library.Port\x12\x0f\n\x07signals\x18\x04 \x03(\t\x12\r\n\x05slots\x18\x05 \x03(\t\"H\n\tBlueprint\x12\r\n\x05group\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x0f\n\x07signals\x18\x04 \x03(\t\x12\r\n\x05slots\x18\x05 \x03(\t\"[\n\x07Library\x12\"\n\x05nodes\x18\x01 \x03(\x0b\x32\x13.mason.library.Node\x12,\n\nblueprints\x18\x02 \x03(\x0b\x32\x18.mason.library.Blueprintb\x06proto3'
)




_PORT = _descriptor.Descriptor(
  name='Port',
  full_name='mason.library.Port',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='mason.library.Port.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='type', full_name='mason.library.Port.type', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='direction', full_name='mason.library.Port.direction', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='sequence', full_name='mason.library.Port.sequence', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='map', full_name='mason.library.Port.map', index=4,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='choices', full_name='mason.library.Port.choices', index=5,
      number=6, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='default', full_name='mason.library.Port.default', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=32,
  serialized_end=150,
)


_NODE = _descriptor.Descriptor(
  name='Node',
  full_name='mason.library.Node',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='group', full_name='mason.library.Node.group', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='name', full_name='mason.library.Node.name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='ports', full_name='mason.library.Node.ports', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='signals', full_name='mason.library.Node.signals', index=3,
      number=4, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='slots', full_name='mason.library.Node.slots', index=4,
      number=5, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=152,
  serialized_end=255,
)


_BLUEPRINT = _descriptor.Descriptor(
  name='Blueprint',
  full_name='mason.library.Blueprint',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='group', full_name='mason.library.Blueprint.group', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='name', full_name='mason.library.Blueprint.name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='signals', full_name='mason.library.Blueprint.signals', index=2,
      number=4, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='slots', full_name='mason.library.Blueprint.slots', index=3,
      number=5, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=257,
  serialized_end=329,
)


_LIBRARY = _descriptor.Descriptor(
  name='Library',
  full_name='mason.library.Library',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='nodes', full_name='mason.library.Library.nodes', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='blueprints', full_name='mason.library.Library.blueprints', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=331,
  serialized_end=422,
)

_NODE.fields_by_name['ports'].message_type = _PORT
_LIBRARY.fields_by_name['nodes'].message_type = _NODE
_LIBRARY.fields_by_name['blueprints'].message_type = _BLUEPRINT
DESCRIPTOR.message_types_by_name['Port'] = _PORT
DESCRIPTOR.message_types_by_name['Node'] = _NODE
DESCRIPTOR.message_types_by_name['Blueprint'] = _BLUEPRINT
DESCRIPTOR.message_types_by_name['Library'] = _LIBRARY
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Port = _reflection.GeneratedProtocolMessageType('Port', (_message.Message,), {
  'DESCRIPTOR' : _PORT,
  '__module__' : 'library_pb2'
  # @@protoc_insertion_point(class_scope:mason.library.Port)
  })
_sym_db.RegisterMessage(Port)

Node = _reflection.GeneratedProtocolMessageType('Node', (_message.Message,), {
  'DESCRIPTOR' : _NODE,
  '__module__' : 'library_pb2'
  # @@protoc_insertion_point(class_scope:mason.library.Node)
  })
_sym_db.RegisterMessage(Node)

Blueprint = _reflection.GeneratedProtocolMessageType('Blueprint', (_message.Message,), {
  'DESCRIPTOR' : _BLUEPRINT,
  '__module__' : 'library_pb2'
  # @@protoc_insertion_point(class_scope:mason.library.Blueprint)
  })
_sym_db.RegisterMessage(Blueprint)

Library = _reflection.GeneratedProtocolMessageType('Library', (_message.Message,), {
  'DESCRIPTOR' : _LIBRARY,
  '__module__' : 'library_pb2'
  # @@protoc_insertion_point(class_scope:mason.library.Library)
  })
_sym_db.RegisterMessage(Library)


# @@protoc_insertion_point(module_scope)
